"""
"Genyal" (c) by Ignacio Slater M.
"Genyal" is licensed under a
Creative Commons Attribution 4.0 International License.
You should have received a copy of the license along with this
work. If not, see <http://creativecommons.org/licenses/by/4.0/>.
"""

from copy import copy
from random import Random
from typing import Any, Callable, Generic, List, Optional, get_args

from genyal.core import DNA, GeneticsError, GenyalCore
from genyal.genotype import GeneFactory
from genyal.operations.crossover import single_point_crossover
from genyal.operations.mutation import simple_mutation


class Individual(GenyalCore, Generic[DNA]):
    """
    Individuals are the basic members of a population.
    On the context of the genetic algorithm engine, a population is a list of individuals from which
    it should be able to pick the fittest one.

    Every individual must have a way to perform a crossover with a partner to give birth to a new
    child; and to mutate according to a given mutation rate.

    Most of the handling process of the individuals will be done by the engine (see:
    genyal.engine.GenyalEngine).
    """
    __fitness: Optional[float]
    __genes: List[DNA]
    __gene_factory: GeneFactory[DNA]
    __mutation_rate: float
    __crossover_strategy: Callable[..., 'Individual[DNA]']
    __mutation_strategy: Callable[..., 'Individual[DNA]']

    def __init__(self, genes=None, mutation_rate=0.01, gene_factory=GeneFactory(),
                 crossover_strategy=single_point_crossover, mutation_strategy=simple_mutation,
                 random_generator=Random()):
        """
        Initializes an individual.
        If no parameters are given to the constructor the individual is created with default
        parameters.

        Args:
            genes:
                A sequence representing the genes of this individual.
                This value defaults to an empty list.
                Genes can be any type of sequence that can be accessed by its index; e.g.,
                strings, lists.
            mutation_rate:
                The probability used by the mutation function to generate a new individual.
            gene_factory:
                A factory to produce the genes of this individual using a generator function (refer
                to: genyal.genotype.GeneFactory).
            crossover_strategy:
                The function to perform the crossover operation.
                Defaults to single_point_crossover (see: genyal.operations.crossover).
            mutation_strategy:
                The function to perform the mutation operation.
                Defaults to simple_mutation (see: genyal.operations.mutation).
            random_generator:
                The random number generator used in the algorithm's operations.
        """
        super(Individual, self).__init__(random_generator)
        self.__fitness = None
        self.__genes = genes if genes is not None else []
        self.__mutation_rate = mutation_rate
        self.__crossover_strategy = crossover_strategy
        self.__mutation_strategy = mutation_strategy
        self.__gene_factory = gene_factory

    @classmethod
    def create(cls, number_of_individuals: int, number_of_genes: int,
               gene_factory: GeneFactory[DNA], mutation_rate: float = 0.01, *args) \
            -> List['Individual[DNA]']:
        """
        Factory method to easily create a population of individuals.

        Args:
            number_of_individuals:
                number of individuals to return.
            number_of_genes:
                the number of genes each individual should have.
            gene_factory:
                a one-argument factory to generate a gene.
            mutation_rate:
                the probability —a number in [0, 1)—  with which an individual will mutate.
        """
        individuals = []
        for _ in range(0, number_of_individuals):
            individual = Individual(gene_factory=gene_factory, mutation_rate=mutation_rate)
            individual.set(number_of_genes, *args)
            individuals.append(individual)
        return individuals

    def compute_fitness_using(self, fitness_function: Callable[..., float], *args):
        """Computes this individual's fitness if it hasn't been computed yet."""
        if not self.__genes:
            raise GeneticsError("The individual should have genes.")
        if self.__fitness is None:
            self.__fitness = fitness_function(self.__genes, *args)

    def set(self, number_of_genes: int, *args):
        """Generate the genes of the individual."""
        for _ in range(0, number_of_genes):
            self.__genes.append(self.__gene_factory.make(*args))

    def crossover(self, partner: 'Individual[DNA]', *args):
        return self.__crossover_strategy(self, partner, *args)

    def mutate(self, *args) -> 'Individual[DNA]':
        return self.__mutation_strategy(self, *args)

    # region : Properties
    @property
    def fitness(self) -> float:
        """The fitness of this individual according to its fitness function."""
        return self.__fitness

    @property
    def genes(self) -> List[DNA]:
        """The genes of this individual"""
        return copy(self.__genes)

    @genes.setter
    def genes(self, new_genes: List[DNA]):
        """Assigns a new set of genes to this individual"""
        self.__genes = list(new_genes)

    @property
    def mutation_rate(self) -> float:
        """The probability with which the genes of the individual will mutate."""
        return self.__mutation_rate

    @mutation_rate.setter
    def mutation_rate(self, rate: float) -> None:
        """The probability with which the genes of the individual will mutate."""
        self.__mutation_rate = rate

    @property
    def crossover_strategy(self) -> Callable[..., 'Individual[DNA]']:
        """The strategy to perform a crossover between two individuals."""
        return self.__crossover_strategy

    @crossover_strategy.setter
    def crossover_strategy(self, strategy: Callable[..., 'Individual[DNA]']):
        """
        The strategy to perform a crossover between two individuals.
        If not set, the strategy is a single-point crossover with a randomly chosen cut point.
        """
        self.__crossover_strategy = strategy

    @property
    def mutation_strategy(self) -> Callable[..., 'Individual[DNA]']:
        """The strategy to perform a mutation over this individual."""
        return self.__mutation_strategy

    @mutation_strategy.setter
    def mutation_strategy(self, strategy: Callable[..., DNA]):
        """The strategy to perform a mutation over this individual."""
        self.__mutation_strategy = strategy

    @property
    def gene_factory(self) -> GeneFactory[DNA]:
        """A factory to generate the genes of this individual."""
        return self.__gene_factory

    @gene_factory.setter
    def gene_factory(self, factory: GeneFactory[DNA]):
        """Sets the factory that generates the genes of this individual."""
        self.__gene_factory = factory

    # endregion

    def __len__(self):
        """The number of genes of this individual"""
        return len(self.__genes)

    def __repr__(self) -> str:
        """An individual is represented by its fitness and its genes."""
        return f"{self.__fitness} - {self.__genes}"

    def __eq__(self, other: Any) -> bool:
        """Two individuals are equal if they have the same fitness"""
        return isinstance(other, Individual) and get_args(self) == get_args(
            other) and self.__fitness == other.__fitness

    def __lt__(self, other: Any) -> bool:
        """Individuals are sorted according to their fitness."""
        return isinstance(other, Individual) and get_args(self) == get_args(
            other) and self.__fitness < other.__fitness

    def __le__(self, other):
        """Individuals are sorted according to their fitness."""
        return isinstance(other, Individual) and get_args(self) == get_args(
            other) and self.__fitness <= other.__fitness

    def __gt__(self, other) -> bool:
        """Individuals are sorted according to their fitness."""
        return isinstance(other, Individual) and get_args(self) == get_args(
            other) and self.__fitness > other.__fitness

    def __ge__(self, other) -> bool:
        """Individuals are sorted according to their fitness."""
        return isinstance(other, Individual) and get_args(self) == get_args(
            other) and self.__fitness >= other.__fitness

    def __copy__(self) -> 'Individual[DNA]':
        """Returns a copy of this individual."""
        return Individual(self.__genes, self.__mutation_rate, self.__gene_factory,
                          self.__crossover_strategy, self.__mutation_strategy,
                          self._random_generator)



"""
"Genyal" (c) by Ignacio Slater M.
"Genyal" is licensed under a
Creative Commons Attribution 4.0 International License.
You should have received a copy of the license along with this
work. If not, see <http://creativecommons.org/licenses/by/4.0/>.
"""
from copy import copy
from random import Random
from typing import List

from genyal.individuals import Individual


def tournament_selection(population: List[Individual], random_generator: Random = Random(),
                         matches: int = 5) -> Individual:
    """
    Selects the fittest from a group of individuals.
    The tournament selection algorithm compares the fitness of two individuals at a time (a match)
    and chooses the one with the best fitness to continue to the next match.

    Args:
        population:
            the collection of individuals to play the tournament.
        random_generator:
            a random number generator to pick individuals for the matches.
        matches:
            the numbers of matches the tournament is going to last.
    Returns:
        The individual who won all the matches (i.e. the one with the highest fitness).
    """
    best_idx = 0
    for _ in range(0, matches):
        candidate_idx = random_generator.randrange(0, len(population))
        if best_idx is None or candidate_idx > best_idx:
            best_idx = candidate_idx
    return copy(population[best_idx])


def default_terminating_function(engine, max_generations=100):
    """
    By default the engine will finish after a certain number of generations have gone through
    evolution.

    Args:
        engine:
            The engine running the genetic algorithm.
        max_generations:
            The number of generations the population will advance.
            If none is given, it defaults to 100 generations.
    Returns:
        True when the engine has evolved the population until the indicated generation.
    """
    return engine.generation >= max_generations



import string
import sys
import unittest
from random import Random, random, randrange

import pytest

from genyal.engine import GenyalEngine
from genyal.genotype import GeneFactory


def match_word_fitness(predicted: list[str], target: str) -> float:
    return sum([predicted[i] == target[i] for i in range(0, len(target))])


def exact_match(engine: GenyalEngine, target: str) -> bool:
    return "".join(engine.fittest.genes) == target


@pytest.mark.repeat(16)
def test_basic_engine(seed: int) -> None:
    basic_engine = GenyalEngine()
    assert basic_engine.fittest is None
    assert basic_engine.generation == 0
    assert len(basic_engine.crossover_args) == 0
    factory = GeneFactory()
    factory.generator = lambda: random()
    random_generator = Random(seed)
    population_size = random_generator.randint(0, 300)
    individual_size = random_generator.randint(0, 100)
    basic_engine.create_population(population_size, individual_size, gene_factory=factory)
    assert basic_engine.fittest is not None
    assert len(basic_engine.population) == population_size
    for individual in basic_engine.population:
        assert len(individual) == individual_size
        assert individual.fitness == 0


@pytest.mark.repeat(32)
def test_word_match_engine(match_word_engine: GenyalEngine, population_size: int,
                           ascii_gene_factory: GeneFactory[str], random_word: str,
                           mutation_rate: float, random_generator: Random, seed: int) -> None:
    match_word_engine.factory_generator_args = (random_generator,)
    match_word_engine.fitness_function_args = (random_word,)
    match_word_engine.create_population(population_size, len(random_word), ascii_gene_factory,
                                        mutation_rate)
    match_word_engine.evolve(random_word)
    assert "".join(match_word_engine.fittest.genes) == random_word, f"Test failed with seed: {seed}"


@pytest.fixture
def match_word_engine(random_generator: Random) -> GenyalEngine:
    return GenyalEngine(random_generator, match_word_fitness, terminating_function=exact_match)


@pytest.fixture()
def ascii_gene_factory() -> GeneFactory[str]:
    return GeneFactory(generator=lambda r: r.choice(string.ascii_lowercase))


@pytest.fixture()
def random_word(random_generator: Random) -> str:
    word = ""
    for _ in range(4, 8):
        word += random_generator.choice(string.ascii_lowercase)
    return word


@pytest.fixture()
def population_size(random_generator: Random) -> int:
    return random_generator.randint(4, 64)


@pytest.fixture()
def mutation_rate(random_generator: Random) -> float:
    return random_generator.random()


@pytest.fixture()
def random_generator(seed: int) -> Random:
    return Random(seed)


@pytest.fixture()
def seed() -> int:
    return randrange(-sys.maxsize, sys.maxsize)


if __name__ == '__main__':
    unittest.main()



"""
"Genyal" (c) by Ignacio Slater M.
"Genyal" is licensed under a
Creative Commons Attribution 4.0 International License.
You should have received a copy of the license along with this
work. If not, see <http://creativecommons.org/licenses/by/4.0/>.
"""
from copy import copy

from genyal.core import GeneticsError


def single_point_crossover(individual, partner, cut_point: int = -1):
    """
    Returns a new child from mixing two individuals using a single-point crossover strategy.
    By default the cut point is selected at random.
    """
    if len(individual) != len(partner):
        raise CrossoverError(
            f"Can't perform a crossover over individuals of different sizes. {len(individual)} != "
            f"{len(partner)}.")
    if cut_point == -1:
        cut_point = individual.random_generator.randrange(0, len(individual.genes))
    child = copy(individual)
    child.random_generator = individual.random_generator
    crossover_genes = individual.genes[:cut_point] + partner.genes[cut_point:]
    child.genes = crossover_genes
    return child


class CrossoverError(GeneticsError):
    """If an error occurs during a crossover operation"""

    def __init__(self, cause: str):
        super(CrossoverError, self).__init__(cause)



"""
"Genyal" (c) by Ignacio Slater M.
"Genyal" is licensed under a
Creative Commons Attribution 4.0 International License.
You should have received a copy of the license along with this
work. If not, see <http://creativecommons.org/licenses/by/4.0/>.
"""
from random import Random
from typing import TypeVar

DNA = TypeVar("DNA")


class GenyalCore:
    """Base for the elements involved in a genetic algorithm's population."""
    _random_generator: Random

    def __init__(self, random_generator: Random):
        """Initializes a genetic algorithm element with a random number generator."""
        self._random_generator = random_generator

    @property
    def random_generator(self) -> Random:
        """The random number generator of this element."""
        return self._random_generator

    @random_generator.setter
    def random_generator(self, new_generator: Random):
        """Sets a new random number generator."""
        self._random_generator = new_generator


class GeneticsError(Exception):
    """When there's an error on the genetic configuration of an element."""
    _cause: str

    def __init__(self, cause):
        super(GeneticsError, self).__init__(cause)
        self._cause = cause

    @property
    def cause(self):
        return self._cause



"""
"Genyal" (c) by Ignacio Slater M.
"Genyal" is licensed under a
Creative Commons Attribution 4.0 International License.
You should have received a copy of the license along with this
work. If not, see <http://creativecommons.org/licenses/by/4.0/>.
"""
from typing import Callable, Generic

from genyal.core import DNA


class GeneFactory(Generic[DNA]):
    """Factory class for creating genes of type DNA."""
    __generator: Callable

    def __init__(self, generator=id):
        """
        Creates a new factory to make genes.
        Args:
            *args:
                In case that the generator function of this factory needs arguments they can be
                supplied by using the constructor.
        """
        self.__generator = generator

    @property
    def generator(self) -> Callable[..., DNA]:
        """The generator is the function that the factory is going to use to create new genes."""
        return self.__generator

    @generator.setter
    def generator(self, generator_function: Callable[..., DNA]) -> None:
        """Gives a new generator function to the factory."""
        self.__generator = generator_function

    def make(self, *args) -> DNA:
        """
        Creates a new gene using the factory's generator function and (if present) the arguments
        given on the constructor.
        """
        return self.__generator(*args)



"""
"Genyal" (c) by Ignacio Slater M.
"Genyal" is licensed under a
Creative Commons Attribution 4.0 International License.
You should have received a copy of the license along with this
work. If not, see <http://creativecommons.org/licenses/by/4.0/>.
"""
import random
import string
import unittest
from random import Random
from typing import List, Tuple

import pytest

from genyal.genotype import GeneFactory
from genyal.individuals import Individual


def test_creation_with_characters(str_gene_factory: GeneFactory[str]):
    individuals = Individual.create(number_of_individuals=100, number_of_genes=10,
                                    gene_factory=str_gene_factory)
    assert len(individuals) == 100
    for individual in individuals:
        assert len(individual.genes) == 10
        for gene in individual.genes:
            assert gene in string.ascii_lowercase


def test_word_guess(str_gene_factory: GeneFactory[str]):
    individuals = Individual.create(number_of_individuals=100000, number_of_genes=3,
                                    gene_factory=str_gene_factory)

    def fitness_function(word: List[str]) -> float:
        return sum([1 if word[i] == "cat"[i] else 0 for i in range(0, 3)])

    fittest_individual = individuals[0]
    for individual in individuals:
        individual.compute_fitness_using(fitness_function)
        if individual.fitness > fittest_individual.fitness:
            fittest_individual = individual
    assert fittest_individual.fitness == 3


def test_crossover(couple: Tuple[Individual, Individual]):
    assert couple[0].crossover(couple[1], 2).genes == list("abfg")
    assert couple[0].crossover(couple[1], 1).genes == list("aefg")
    assert couple[0].crossover(couple[1], 0).genes == list("defg")


def test_mutation():
    gene_factory = GeneFactory[str]()
    gene_factory.generator = lambda: Random(8000).choice(string.ascii_lowercase)
    individual = Individual()
    individual.random_generator = Random(8000)
    individual.genes = "abcd"
    individual.gene_factory = gene_factory
    individual.mutation_rate = 0.5
    mutated_i = individual.mutate()
    assert mutated_i.genes == list("sbsd")


@pytest.fixture()
def couple() -> Tuple[Individual, Individual]:
    c = (Individual(), Individual())
    c[0].genes = "abcd"
    c[1].genes = "defg"
    return c


@pytest.fixture()
def str_gene_factory(rng: Random) -> GeneFactory[str]:
    gene_factory = GeneFactory()
    gene_factory.generator = lambda: rng.choice(string.ascii_lowercase)
    return gene_factory


@pytest.fixture
def rng(seed: int) -> Random:
    return Random(seed)


@pytest.fixture
def seed() -> int:
    return random.randint(-8000, 8000)


if __name__ == '__main__':
    unittest.main()



"""
"Genyal" (c) by Ignacio Slater M.
"Genyal" is licensed under a
Creative Commons Attribution 4.0 International License.
You should have received a copy of the license along with this
work. If not, see <http://creativecommons.org/licenses/by/4.0/>.
"""
import string
import sys
import unittest
from random import Random, randrange

import pytest

from genyal.genotype import GeneFactory


@pytest.mark.repeat(8)
def test_basic_factory(seed: int) -> None:
    factory = GeneFactory[float]()
    gene = Random(seed).random()
    assert factory.generator(gene) == id(gene), f"Test failed with seed: {seed}"


@pytest.mark.repeat(8)
def test_binary_factory(seed: int) -> None:
    rand_generator = Random(seed)
    factory = GeneFactory(generator=lambda rng: rng.random() > 0.5)
    assert factory.make(Random(seed)) == (
            rand_generator.random() > 0.5), "Test failed with seed: {seed}"


@pytest.mark.repeat(32)
def test_ascii_factory(seed: int) -> None:
    gene = Random(seed).choice(string.ascii_lowercase)
    factory = GeneFactory()
    assert factory.make(gene) == id(gene), f"Test failed with seed: {seed}"
    factory.generator = lambda: Random(seed).choice(string.ascii_lowercase)
    assert factory.make() == gene, f"Test failed with seed: {seed}"


@pytest.fixture()
def seed() -> int:
    return randrange(-sys.maxsize, sys.maxsize)


if __name__ == '__main__':
    unittest.main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="genyal",  # Replace with your own username
    version="0.2.0.1",
    author="Ignacio Slater Muñoz",
    author_email="islaterm@gmail.com",
    description="A framework for genetic algorithms in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/islaterm/genyal",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        # "License :: OSI Approved :: Creative Commons Attribution 4.0 International License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
)



"""
"Genyal" (c) by Ignacio Slater M.
"Genyal" is licensed under a
Creative Commons Attribution 4.0 International License.
You should have received a copy of the license along with this
work. If not, see <http://creativecommons.org/licenses/by/4.0/>.
"""
from random import Random
from typing import Any, Callable, List, Optional, Tuple

from genyal.core import GenyalCore
from genyal.genotype import GeneFactory
from genyal.individuals import Individual
from genyal.operations.evolution import default_terminating_function, tournament_selection


class GenyalEngine(GenyalCore):
    """
    The engine is the main component of Genyal.
    This class is in charge of creating, maintaining and evolving a population.
    """
    __fitness_function_args: Tuple
    __factory_generator_args: Tuple
    __crossover_args: Tuple
    __fitness_function: Callable[[List[Any]], float]
    __fittest: Optional[Individual]
    __generations: int
    __mutation_args: List[Any]
    __population: List[Individual]
    __selection_args: List[Any]
    __selection_strategy: Callable[..., Individual]
    __terminating_function: Callable[..., bool]

    def __init__(self, random_generator: Random = Random(),
                 fitness_function: Callable[..., float] = lambda _: 0,
                 selection_strategy=tournament_selection,
                 terminating_function=default_terminating_function):
        """
        Initializes the values of the engine.

        Args:
            random_generator:
                The random number generator used by the engine.
            fitness_function:
                The function to calculate the fitness of the population's individuals.
                If none given, the default function returns 0 for any individual.
            selection_strategy:
                The strategy to select the individuals that will participate in the crossover.
            terminating_function:
                The function that will decide when to stop the evolution.
        """
        super(GenyalEngine, self).__init__(random_generator)
        self.__population = []
        self.__fitness_function = fitness_function
        self.__fitness_function_args = ()
        self.__fittest = None
        self.__selection_strategy = selection_strategy
        self.__selection_args = []
        self.__crossover_args = ()
        self.__mutation_args = []
        self.__terminating_function = terminating_function
        self.__generations = 0
        self.__factory_generator_args = ()

    def create_population(self, population_size: int, individual_size: int,
                          gene_factory: GeneFactory, mutation_rate=0.01):
        """
        Creates a new population for the engine.
        The new population is then sorted according to the individual's fitness

        Args:
            mutation_rate:
            population_size:
                The number of individuals of the population.
            individual_size:
                The number of genes of each member.
            gene_factory:
                The factory to create the genes of each individual
        """
        self.__population = Individual.create(population_size, individual_size, gene_factory,
                                              mutation_rate, *self.__factory_generator_args)
        for member in self.__population:
            member.compute_fitness_using(self.__fitness_function, *self.__fitness_function_args)
        self.__population.sort()
        self.__fittest = self.__population[-1]

    def evolve(self, *args):
        """
        Evolves the population until the condition given by the terminating function is met.
        By default, the population will evolve until it reaches 100 generations.

        Args:
            *args:
                The arguments passed to the terminating function.
        """
        while not self.__terminating_function(self, *args):
            new_population = []
            for _ in range(0, len(self.__population)):
                child = self.__create_offspring()
                new_population.append(child)
            new_population.sort()
            self.__population = new_population
            self.__fittest = new_population[-1]
            self.__generations += 1

    def crossover(self, partner_a: Individual, partner_b: Individual, *args) -> Individual:
        """Performs a crossover between two individuals and returns the offspring."""
        partner_a.random_generator = self._random_generator
        return partner_a.crossover(partner_b, *args)

    def mutate(self, individual: Individual, *args) -> Individual:
        """Mutates an individual and returns the result of the mutation."""
        individual.random_generator = self.random_generator
        return individual.mutate(*args)

    def __create_offspring(self):
        """
        Creates an offspring from a couple.
        The partners are selected from the population and the offspring is obtained via crossover
        and mutation.
        """
        partner_a = self.__selection_strategy(self.__population, self._random_generator,
                                              *self.__selection_args)
        partner_b = self.__selection_strategy(self.__population, self._random_generator,
                                              *self.__selection_args)
        child = self.mutate(self.crossover(partner_a, partner_b, *self.__crossover_args),
                            self._random_generator,
                            *self.__mutation_args)
        child.compute_fitness_using(self.__fitness_function, *self.__fitness_function_args)
        return child

    @property
    def population(self) -> List[Individual]:
        """The individuals of the current generation."""
        return self.__population

    @property
    def generation(self) -> int:
        """The number of generations the population has evolved."""
        return self.__generations

    @property
    def fittest(self):
        """The individual with the greatest fitness from the population"""
        return self.__fittest

    @property
    def crossover_args(self) -> Tuple:
        """A tuple with extra arguments to be passed to the crossover operation."""
        return self.__crossover_args

    @crossover_args.setter
    def crossover_args(self, args):
        """Sets the arguments needed by the crossover operation."""
        self.__crossover_args = args

    @property
    def factory_generator_args(self) -> Tuple:
        return self.__factory_generator_args

    @factory_generator_args.setter
    def factory_generator_args(self, args: Tuple) -> None:
        self.__factory_generator_args = args

    @property
    def fitness_function_args(self) -> Tuple:
        return self.__fitness_function_args

    @fitness_function_args.setter
    def fitness_function_args(self, args: Tuple) -> None:
        self.__fitness_function_args = args



"""
"Genyal" (c) by Ignacio Slater M.
"Genyal" is licensed under a
Creative Commons Attribution 4.0 International License.
You should have received a copy of the license along with this
work. If not, see <http://creativecommons.org/licenses/by/4.0/>.
"""

from copy import copy


def simple_mutation(original_individual, *args):
    """
    Returns a new individual resulting from mutating the original with a given mutation rate.
    The second argument is a placeholder.
    """
    new_individual = copy(original_individual)
    new_individual.genes = [
        gene if new_individual.random_generator.random() <= new_individual.mutation_rate
        else new_individual.gene_factory.make(*args)
        for gene in new_individual.genes]
    return new_individual
